#include "SubmatClass.h"


SubmatClass_t::SubmatClass_t() {
}

SubmatClass_t::SubmatClass_t(AAAlphabet& ssealph) {
    /* define aa alphabet order and create index map*/
    alphabet = ssealph.getAlphabet();
    assert(alphabet == "CEH");
           
    for (int i = 0; i < alphabet.length(); i++) {
        aa2idx.insert (pair<char,int>(alphabet[i],i));
    }

    string    S_filepath = "data/SSTSUM_S_double_precision.txt";
    string    D_filepath = "data/SSTSUM_D_double_precision.txt";
    string invS_filepath = "data/SSTSUM_Sinv_double_precision.txt";

    //assumes reading an alphabet.size() x alphabet.size() matrices below.
    S        = readMatrix(S_filepath);
    D        = readMatrix(D_filepath);
    invS     = readMatrix(invS_filepath);
   
    generateStationaryDistribution();
    generateConditionalMatrices();
    generateJointMatrices();
    Mpow1 = cMat[1];

    assert(abs(getExpectedChange() - 0.01) < 1e-3);

    fsaobj = FSAClass_t(ssealph.size());
}

vector<vector<double> > SubmatClass_t::readMatrix(string fn) {
    ifstream infile(fn.c_str(),ios::in);
    assert(infile);
    vector<vector<double> > mat;
    size_t nRows = alphabet.size();
    size_t nCols = nRows;
    for (size_t i = 0; i < nRows; i++) {
        vector<double> row;
        double tmp;
        for (size_t j = 0; j < nCols; j++) {
            infile >> tmp;
            row.push_back(tmp);
        }
        mat.push_back(row);
    }
    infile.close();
    return mat;
}

void SubmatClass_t::generateStationaryDistribution() {
    //Find eigvec in S corresp. to eigvalue=1 in D.
    //Considering floating point rep, instead find the index of largest eigvalue
    //in D and then read the corresp. col vec in S, and normalize.
    double evaloneidx = 0;
    double maxval = D[0][0];
    for( int i = 1; i < alphabet.size(); i++) {
        if (D[i][i] > maxval) {
            maxval = D[i][i];
            evaloneidx = i;
        }
    }
    //read evaloneindx col in S matrix
    vector<double> evec;
    double sum = 0.0;
    for (int i =0; i < alphabet.size(); i++) {
        evec.push_back(S[i][evaloneidx]);
        sum += S[i][evaloneidx];
    }

    //normalize evec and write to stationarydist
    double normsum = 0.0;
    for (int i =0; i < alphabet.size(); i++) {
        double normval = (double)evec[i]/(double)sum;
        stationarydist.push_back(normval);
        normsum += normval;
    }
    assert(stationarydist.size() == alphabet.size());
    assert(normsum > 0.999999 && normsum < 1.000001);
    //check if all elems lie between [0,1]
    for (int i = 0; i < alphabet.size(); i++) {
        assert(stationarydist[i] >=0 && stationarydist[i] <=1);
    }
}

void SubmatClass_t::generateConditionalMatrices() {
    assert( cMat.size()== 0); // sanity check;
    for (int N = 0; N <= NMATRICES; N++) {
        vector<vector<double> > DpowN = D;
        for (size_t i = 0;  i < D.size(); i++) {
            DpowN[i][i] = pow(D[i][i],N);
        }
        vector<vector<double> > mat =  matrixMultiply(S,matrixMultiply(DpowN,invS));
        //normalize mat.
        normalizeConditionalMat(mat);
        cMat.push_back(mat);
    }
    assert( cMat.size()== NMATRICES+1); // matrices for N=0 to N=NMATRICES;
}

vector<vector<double>> SubmatClass_t::generateConditionalMatrices(int N) {
  
    vector<vector<double> > DpowN = D;
    for (size_t i = 0;  i < D.size(); i++) {
        DpowN[i][i] = pow(D[i][i],N);
    }
    vector<vector<double> > mat =  matrixMultiply(S,matrixMultiply(DpowN,invS));
    //normalize mat.
    normalizeConditionalMat(mat);

    return mat;
}

void SubmatClass_t::generateJointMatrices() {
    assert( cMat.size()== NMATRICES+1 && jMat.size() == 0); //sanity check
    
    for (int N = 0; N <=NMATRICES; N++) {

        vector<vector<double> > mat;
        //alloc memory for mat
        for (int i =0; i < alphabet.size(); i++) {
            mat.push_back(vector<double>(alphabet.size(),0.0));
        }
        //compute joint using stationary and conditional
        for (int i = 0; i < alphabet.size(); i++) {
            for (int j = i; j < alphabet.size(); j++) {
                mat[i][j] = (double)(stationarydist[j]*cMat[N][i][j] 
                             +       stationarydist[i]*cMat[N][j][i])
                             /(double)2;
                if (i != j) mat[j][i] = mat[i][j];
            }
        }
        normalizeJointMat(mat);
        jMat.push_back(mat);
    }
    assert( jMat.size()== NMATRICES+1); // matrices for N=0 to N=NMATRICES;
}


void SubmatClass_t::regenerateJointMatricesUsingNullModelEstimates(NullModel& nullMdl) {
    cout << "!!!!!@@@@@#####\n";
    assert( cMat.size()== NMATRICES+1 && jMat.size() == NMATRICES+1); //sanity check
    
    for (int N = 0; N <=NMATRICES; N++) {
        //compute joint using stationary and conditional
        for (int i = 0; i < alphabet.size(); i++) {
            for (int j = i; j < alphabet.size(); j++) {
                string char_i(1,alphabet[i]);
                string char_j(1,alphabet[j]);

                jMat[N][i][j] = (double)(nullMdl.pr(char_j)*cMat[N][i][j] 
                             +           nullMdl.pr(char_i)*cMat[N][j][i])
                             /(double)2;
                if (i != j) jMat[j][i] = jMat[i][j];
            }
        }
        normalizeJointMat(jMat[N]);
    }
    assert( jMat.size()== NMATRICES+1); // matrices for N=0 to N=NMATRICES;
}


//joint pr of x,y at  N
double SubmatClass_t::jointpr(char x, char y, int N) {
    map<char,int>::iterator it;
    it = aa2idx.find(x);
    assert(it != aa2idx.end());
    it = aa2idx.find(y);
    assert(it != aa2idx.end());
    return jMat[N][aa2idx[x]][aa2idx[y]];
}

//conditional pr x|y at N
double SubmatClass_t::condlpr(char x, char y, int N) {
    map<char,int>::iterator it;
    it = aa2idx.find(x);
    assert(it != aa2idx.end());
    it = aa2idx.find(y);
    assert(it != aa2idx.end());
    return cMat[N][aa2idx[x]][aa2idx[y]];
}


string SubmatClass_t::getAlphabet() {
    return alphabet;
}

double SubmatClass_t::getS(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return S[x][y];
}

double SubmatClass_t::getSinv(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return invS[x][y];
}

double SubmatClass_t::getD(int x,int y) {
    static int n = alphabet.length();
    assert(x < n && x >= 0);
    assert(y < n && y >= 0);
    return D[x][y];
}

int SubmatClass_t::getIndx(char x) {
    assert(alphabet.find(x)!=string::npos);
    return aa2idx[x];
}

void SubmatClass_t::normalizeConditionalMat(vector<vector<double> > &M) {
    /* assumes input is a conditional probability matrix, and normalizes
     * by ensuring all cols add to 1 up to double precision
     * */
    int size = alphabet.length();

    /* When N is small (eg N=1; note: M input is an instance of MpowN),
     * some elements of M approach 0 from the -ve side due to numerical 
     * issues. So, first convert values <0 to 0 */
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            if (M[i][j] < 0) M[i][j] = 0.0;
        }
    }

    //find col sums
    vector<double> colsum;
    for (int j = 0;  j < size; j++) {
        double tmpsum = 0;
        for (int i = 0;  i < size; i++) {
            tmpsum += M[i][j];
        }
        colsum.push_back(tmpsum);
    }

    //normalize
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            M[i][j] /= colsum[j];
        }
    }
}

void SubmatClass_t::normalizeJointMat(vector<vector<double> > &M) {
    /* assumes input is a joint probability matrix, and normalizes
     * by ensuring all elements over the matrix add to 1 to double precision
     * */
    int size = alphabet.length();

    //ensure no elem is <0
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            assert(M[i][j] >= 0);
        }
    }

    //find col sums
    double allsum = 0.0;
    for (int j = 0;  j < size; j++) {
        for (int i = 0;  i < size; i++) {
            allsum += M[i][j];
        }
    }

    //normalize
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            M[i][j] /= allsum;
        }
    }
}


void SubmatClass_t::printStationaryDist() {
    cout << "--Printing stationary distribution of the matrix:"<< endl;
    for (int i = 0;  i < alphabet.size(); i++) {
        cout <<  alphabet[i] << " " << setw(10) << setprecision(7) << stationarydist[i] << endl;
    }
    cout << "---\n";
}

void SubmatClass_t::printMatpowN(string mattype,int N) { // mattype = {"conditional", "joint"}
    assert(N>=0 && N <=NMATRICES);
    cout << "--Printing details of " << mattype << "Mat(" << N << ")"<< endl;



    int size = alphabet.length();
    //find row sums
    vector<double> rowsum;
    double allsum = 0.0;
    for (int i = 0;  i < size; i++) {
        double tmpsum = 0;
        for (int j = 0;  j < size; j++) {
            if (mattype.compare("conditional") == 0) {
                tmpsum += cMat[N][i][j];
                allsum += cMat[N][i][j];
            }
            else if (mattype.compare("joint") == 0) {
                tmpsum += jMat[N][i][j];
                allsum += jMat[N][i][j];
            }
        }
        rowsum.push_back(tmpsum);
    }
    //find col sums
    vector<double> colsum;
    for (int j = 0;  j < size; j++) {
        double tmpsum = 0;
        for (int i = 0;  i < size; i++) {
            if (mattype.compare("conditional") == 0) {
                tmpsum += cMat[N][i][j];
            }
            else if (mattype.compare("joint") == 0) {
                tmpsum += jMat[N][i][j];
            }
        }
        colsum.push_back(tmpsum);
    }

    int w = 16, p = 12;
    cout << fixed;
    //alphabet row
    cout << setw(w) << " ";
    for (int i = 0;  i < size; i++) {
        cout << setw(w) << alphabet[i];
    }
    cout << " | ";
    cout << setw(w) << "rowsum";
    cout << endl;

    //matrix rows
    for (int i = 0;  i < size; i++) {
        cout << setw(w) << alphabet[i];
        for (int j = 0;  j < size; j++) {
            if (mattype.compare("conditional") == 0) {
                cout << setw(w) << setprecision(p) << cMat[N][i][j];
            }
            else if (mattype.compare("joint") == 0) {
                cout << setw(w) << setprecision(p) << jMat[N][i][j];
            }
        }
        cout << " | ";
        cout << setw(w) << setprecision(p) << rowsum[i];
        cout << endl;
    }
    //col sum line
    cout << "------";
    cout << endl;
    cout << setw(w) << "colsum";
    for (int j = 0;  j < size; j++) {
        cout << setw(w) << setprecision(p) << colsum[j];
    }
    cout << setw(w) << setprecision(p) << allsum;
    cout << endl;
    cout << endl;

}

void SubmatClass_t::scaleMatrix(){
    double x = (double)0.01000000000000; // standard expected change
	double x_curr = getExpectedChange();

    cout << x_curr << endl;
    if(abs(getExpectedChange() - 0.01) < 1e-10)
        return;

	double scaling_factor = x/x_curr; 
    
    assert(D.size() == Mpow1.size());
    assert(D.size() > 0);
    vector<vector<double>> prevd = D;
    for(int j = 0; j < D.size(); j++){
        if(D[j][j] < 0)
            D[j][j] = 0.0;
		this->D[j][j] = pow(D[j][j], scaling_factor); 
	}
    Mpow1 = generateConditionalMatrices(1);

    double newExpChange = getExpectedChange();

    if(!(abs(newExpChange - 0.01) < 1e-10)) {
        cout << newExpChange << endl;
    }
    
    assert(abs(newExpChange - 0.01) < 1e-10);
}

double SubmatClass_t::getExpectedChange(){
	double Pr_no_change = 0.0; 

    assert(stationarydist.size() == Mpow1.size());
    assert(Mpow1.size() > 0);
	for(int j = 0; j < Mpow1.size(); j++){
		Pr_no_change += (this->stationarydist[j] * this->Mpow1[j][j]); 
	}
	return ((double)1.0 - Pr_no_change); 
}

double SubmatClass_t::getprMM(int N) {
    return fsaobj.getprMM(N);
}
double SubmatClass_t::getprII(int N) {
    return fsaobj.getprII(N);
}
double SubmatClass_t::getprIM(int N) {
    return fsaobj.getprIM(N);
}
